/*
  ProgEx33 -- C++ Version Of String::Parse
  
  I remember being really thrilled when I discovered PowerBASIC's ParseCount -
  Parse combination.  Its language elements like that that really help applic-
  ation programmers get things done.  And that is the sort of language element
  which C always lacked - C's strtok() function nonwithstanding.  So when I
  built my C++ String Class its something I wanted to include for sure.  Before
  looking at a C++ String Class implementation, lets first look at how Parse
  works in PowerBASIC so we're all on the same page. Lets imagine we have my
  name to Parse with my first, middle and last names seperated ny commas...
  
  'First PowerBASIC Version
  #Compile Exe
  #Dim All

  Function PBMain() As Long
    Local strArr() As String
    Local strName As String
    Register i As Long
  
    strName="Frederick, John, Harris"
    Redim strArr(ParseCount(strName,",")-1)
    Parse strName, strArr(), ","
    Print " i            strArr(i)"
    Print "========================"
    For i=0 To UBound(strArr,1)
      Print i, LTrim$(strArr(i))
    Next i
    Erase strArr()
    Waitkey$

    PBMain=0
  End Function

  ' i            strArr(i)
  '========================
  ' 0            Frederick
  ' 1            John
  ' 2            Harris
  
  
  //Then C++ Version With My String Class
  #include <stdio.h>
  #include "Strings.h"

  int main(void)
  {
   String s1("Frederick, John, Harris");
   unsigned int iCount;
   String* pStr=NULL;   
 
   iCount=s1.ParseCount(',');
   pStr=new String[iCount];
   s1.Parse(pStr,',');
   printf("&pStr[i]\tpStr[i].lpStr()\t\tpStr[i].lpStr()\n");
   printf("=======================================================\n");
   for(unsigned int i=0; i<iCount; i++)
   {
       pStr[i].LTrim();
       printf("%u\t\t%u\t\t\t%s\n",(unsigned int)&pStr[i],(unsigned int)pStr[i].lpStr(),pStr[i].lpStr());
   }    
   delete [] pStr; 
   getchar();   
    
   return 0;   
  }

  The PowerBASIC version comes out to 21 lines and the C++ version to 24.  
  The comparison is certainly interesting though.  The above C++ code isn't
  exactly like the compilable version below because in that version I put all
  the processing code in a procedure named DoProcessing() so that the 
  destructors for all the various Strings would get called and print diagnostic
  output statements for your examination.  As I mentioned several Program
  Examples back, destructors for Strings created in main() are sometimes not
  called until after the console has been closed and main() has exited.  I
  wanted to avoid that. 
  
  In the past several String Class Program Examples I've only added a new 
  feature or two in each new example.  At that rate it will take me forever to
  get through all this.  So in this ProgEx I'm going to just provide my whole
  Ansi version of my String Class as it now stands.  However, I'll provide some
  derivation information on the development of Parse because its so useful. The
  other members that emulate BASIC String functions such as Left, Right, Mid,
  InStr, Trim, etc., you can probably figure out from the source code or just
  use it in your programs.
  
  I set up my C++ Parse as a String Class member function that returns void and
  needs to have the token that will be parsed against and the block of memory
  that will manage the parsed Strings as parameters.  So the initial setup will
  involve a pointer to a String Object declaration such as this...
  
  String* pStr=NULL;
  
  This is somewhat analagous to the declaration of the strArr() variable in the
  PowerBASIC program...
  
  Local strArr() As String
  
  Now that C++ pointer to String object declaration will need to point to memory
  where some as yet undetermined number of String objects will lie.  The first 
  order of business then is to determine the number of Strings in the main 
  String that will need to be parsed.  I have a ParseCount member in my String
  Class that returns that number.  Its implementation is rather trieval in that
  it simply runs through a String object with a char* counting delimiters.  The
  number of Strings will be the number of delimiters counted plus one.  Its use
  is as follows...
  
  String s1("Frederick, John, Harris");
  
  iCount=s1.ParseCount(',');
  pStr=new String[iCount]; 
  
  Since there are three Strings in the main String we'll need to allocate room
  for three Strings with C++'s new operator.  These would be referenced then as
  
  pStr[0], pStr[1], and pStr[2]
  
  The entire block of memory allocated here will be 24 bytes because each String
  comprises 8 bytes; 4 bytes for private member pStrBuffer and 4 more for 
  private member iAllowableCharacterCount.  For sure the parsed Strings don't go
  here.  What goes here is the String object 'overhead' that manages each String
  once it has been parsed.
  
  At this point I believe my narritive should jump to the output from running 
  the program.  The first String Constructor call occurs when String s1 is
  declared and initialized simultaneously with a char string...
  
  String s1("Frederick, John, Harris");
  
  Following that is the call to ParseCount and the return of 3.  Then we have
  three more Uninitialized Constructor calls generated by this statement...
  
  pStr=new String[iCount];
  
  If you look at the value of the 'this' pointer you'll see it starts at 4009100
  and increments up to 4009116 in 8 byte increments for the reasons just 
  described above.  You can think of this memory as the machinery that manages
  the Strings.  The Strings themselves will be located at the numbers held in
  the pStrBuffer members, and as you can see that memory is scattered about a
  good bit.  When we finally make the call to the Parse member...
  
  s1.Parse(pStr,',');
  
  ...our output jumps inside that function reproduced here without the printf
  statements...
   
  
  void String::Parse(String* pStr, char delimiter)
  {
   unsigned int i=0;
   char* pBuffer=0;
   char* c;
   char* p;

   pBuffer=new char[this->LenStr()+1];
   if(pBuffer)
   {
      p=pBuffer;
      c=this->pStrBuffer;
      while(*c)
      {
         if(*c==delimiter)   //We've hit a delimiter and we know that up to this
         {                   //point a null was copied after each non-delimiter
            pStr[i]=pBuffer; //char.  So we should have a null terminated String 
            p=pBuffer;       //at pBuffer that can be directly assigned to one
            i++;             //of the pStr[i] Strings passed in.  Of course i is
         }                   //incremented after each assignment so we're in 
         else                //high cotton.
         {
            *p=*c;  //What's pointed at by c isn't a delimiter so copy the char
            p++;    //p=pBuffer and then increment p, i.e., p++.  Then store a
            *p=0;   //null byte at p which is now directly after the last char.
         }
         c++;
      }
      pStr[i]=pBuffer;
      delete [] pBuffer;
   }
  } 
  
  The 1st thing my Parse does is allocate a temporary local buffer to slop
  around in.  I make it as big as the original String to be parsed.  Then
  I simply start moving through the String to be parsed pointed to by pStrBuffer
  one byte at a time using a while loop construct.  If the byte pointed at by c
  ( *c ) is not a delimiter, I simply copy that byte to my slop buffer pBuffer.
  Then I increment p and put a null byte after the last character copied because
  the next byte at c could very well be a delimiter, in which case the string
  in pBuffer up to the null will have to be assigned to one of the pStr[i]
  passed in.  
  
  So the loop just keeps doing that, i.e., copying bytes to pBuffer and placing 
  a null after each byte, until a delimiter is 'hit'.  When the delimiter is
  hit we're in the upper part of the if and the String accumulated in pBuffer
  is simply assigned to the correct pStr[i] parameter as determined by the i
  counter variable which itself gets incremented after each assignment.
  Interestingly, you can see how the pStr[i]=pBuffer assignment triggered the
  operator= member.  Its all there in the output from within Parse. You can see
  where each of the three Strings were parsed out, and you can see the results
  and all the implicated addresses back in DoProcessing.  Note finally how
  the destructors on all the various Strings were automatically called when
  DoProcessing() terminated.
*/

#include <stdio.h>
#include "Strings.h"

void DoProcessing(void)
{
 String s1("Frederick, John, Harris");
 unsigned int iCount;
 String* pStr=NULL;   
 
 iCount=s1.ParseCount(',');
 printf("iCount = %u\n\n", iCount);
 pStr=new String[iCount];
 printf("pStr   = %u\n\n", (unsigned int)pStr);
 s1.Parse(pStr,',');
 printf("&pStr[i]\tpStr[i].lpStr()\t\tpStr[i].lpStr()\n");
 printf("=======================================================\n");
 for(unsigned int i=0; i<iCount; i++)
 {
     pStr[i].LTrim();
     printf
     (
      "%u\t\t%u\t\t\t%s\n",
      (unsigned int)&pStr[i],(unsigned int)pStr[i].lpStr(),pStr[i].lpStr()
     );
 }    
 delete [] pStr; 
 return;       
}    

int main(void)
{
 DoProcessing();
 getchar();   
    
 return 0;   
}


/*
Entering String(const char* pStr)  //Constructor: Initializes with char*
  this = 2293552
  pStr = Frederick, John, Harris
  pStrBuffer = 4009056  Frederick, John, Harris
Leaving String(const char* pStr)

iCount = 3

Entering String()  Uninitialized Constructor
  this = 4009100
  pStrBuffer = 4008952
Leaving Uninitialized Constructor

Entering String()  Uninitialized Constructor
  this = 4009108
  pStrBuffer = 4009136
Leaving Uninitialized Constructor

Entering String()  Uninitialized Constructor
  this = 4009116
  pStrBuffer = 4009152
Leaving Uninitialized Constructor

pStr   = 4009100

Entering Parse(String* pStr, char delimiter)
  pBuffer = 4009168
  Entering String::operator=(const char* pStr)
    pStr = 4009200      Frederick
  Leaving String::operator=(const char* pStr)

  Assigned pStr[0] In Parse()
  pStr[0]  = Frederick
  pStr[0]  = 4009200
  &pStr[i] = 4009100

  Entering String::operator=(const char* pStr)
    pStr = 4009136       John
  Leaving String::operator=(const char* pStr)

  Assigned pStr[1] In Parse()
  pStr[1]  =  John
  pStr[1]  = 4009136
  &pStr[i] = 4009108

  Entering String::operator=(const char* pStr)
    pStr = 4009224       Harris
  Leaving String::operator=(const char* pStr)

  Assigned pStr[2] In Parse()
  pStr[2]  =  Harris
  pStr[2]  = 4009224
  &pStr[i] = 4009116
Leaving Parse(String* pStr, char delimiter)

&pStr[i]        pStr[i].lpStr()         pStr[i].lpStr()
=======================================================
4009100         4009200                 Frederick
4009108         4009136                 John
4009116         4009224                 Harris

Entering String Destructor!
  this       = 4009116
  pStrBuffer = 4009224  Harris
Leaving String Destructor!


Entering String Destructor!
  this       = 4009108
  pStrBuffer = 4009136  John
Leaving String Destructor!


Entering String Destructor!
  this       = 4009100
  pStrBuffer = 4009200  Frederick
Leaving String Destructor!


Entering String Destructor!
  this       = 2293552
  pStrBuffer = 4009056  Frederick, John, Harris
Leaving String Destructor!
*/
